#!/usr/bin/perl -T
##
use strict;
use Getopt::Long;
use Sys::Syslog qw(:standard :macros);
use Socket qw( inet_aton );

$ENV{'PATH'}='/sbin:/usr/bin:/usr/sbin';  # secure enough?
delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

# List is from RQF0775810 - 2017-06-14
my $remote_nets = [
    '188.185.2.0/23',
    '188.185.16.0/18', # 188.185.16.0/21
                       # 188.185.24.0/22
                       # 188.185.28.0/22
                       # 188.185.32.0/20
                       # 188.185.48.0/20 (.. 188.185.63.255)
    '188.185.128.0/17',
    '10.48.0.0/12',    # 10.48 - 10.63
    '192.168.70.0/23', # 192.168.70 - 192.168.71
    ];

my $opt_verbose=0;
my $opt_noaction=0;
my $opt_help =0;

sub get_my_own_ip() {
    my $ipaddr ='';
    my $iproute=`ip route list 2>/dev/null`;
    $iproute=`busybox iproute list 2>/dev/null` unless ($iproute);
    if ($iproute =~ /^default\s+via\s+([0-9.]+)\s+dev\s+([a-z0-9A-Z-]+)/m) {
	my $defdev = $2;
	print "#INFO: default network device: $defdev\n" if ($opt_verbose);
	my $ipaddr_show=`ip addr show dev '$defdev' 2>/dev/null`;
	$ipaddr_show=`busybox ipaddr show dev '$defdev' 2>/dev/null` unless ($ipaddr_show);
	if($ipaddr_show =~ /\s+inet\s+\S+\s+brd\s+(\d+\.\d+\.\d+\.\d+)\s+scope/s) {
	    $ipaddr = $1;
	    print "#INFO: my IP: $ipaddr\n" if ($opt_verbose);
	} else {
	    print "#WARN: default network device ' $defdev' has no IPv4 address\n";
	}
    } else {
	print "#WARN: no default network device found\n";
	syslog(LOG_WARNING, "no default network device found");
    }
    return $ipaddr;
}

# perhaps could 'translate' into raw numbers once, but these operations
# are really fast.
sub ip_in_cidrs ($$) {
    my $ip = shift;
    my $cidrs = shift;
    my $ip_l = unpack( 'N', inet_aton($ip));
    foreach my $cidr (@{$cidrs}) {
	if ($cidr =~ m:(^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/(\d{1,2})$: ) {
	    my $start_l = unpack( 'N', inet_aton($1));
	    my $bits = $2;
	    my $mask = -1<<(32-$bits);
	    $start_l &= $mask;

	    if( ($ip_l & $mask)==$start_l )	{
		return( 1 );
	    }
	} else {
	    die("invalid CIDR syntax for $cidr");
	}
    }
    return 0;
}
sub get_fileservers($) {
    my $cell = shift;
    my @fileservers_raw=`vos listaddrs -cell '$cell' -noresolve -noauth`;
    my @fileservers;
    for my $s (@fileservers_raw) {
	if ($s =~ /^(\d+\.\d+\.\d+\.\d+)$/) {
	    push(@fileservers, $1);
	}
    }
    return @fileservers;
}

sub get_mycell() {
    my $mycell = '';
    my $mycell_raw=`fs wscell`;
    if($mycell_raw =~ m/belongs to cell .([a-z.]+)./s) {
	$mycell = $1;
    }
    print "#INFO: this AFS client is in cell $mycell\n" if $opt_verbose;
    return $mycell;
}

sub get_dbservers( $ ) {
    my $mycell=shift;
    my $cellservdb='NoCellServDBFoundInStandardPlaces';
    for my $f ('/usr/vice/etc/CellServDB', '/etc/openafs/CellServDB') {
	if ( -r $f) {
	    $cellservdb = $f;
	    last;
	}
    }
    if (! open(CELLSERVDB, "<$cellservdb")) {
	print "#ERROR: could not open $cellservdb:$!";
	return;
    }
    my @allcells = <CELLSERVDB>; # small enough for 'slurp'
    close(CELLSERVDB);
    my $foundcell = 0;
    my @dbservers;
    for my $line (@allcells) {
	if ($line =~ m/^>$mycell\s+/) {   # start of 'our' block
	    $foundcell =1;
	    next;
	} elsif ($foundcell and $line =~ m/^\[?(\d+\.\d+\.\d+\.\d+)\]?\s+/) {
	    push(@dbservers, $1);
	    next;
	} elsif ($foundcell and $line =~ m/^>/) { # next block
	    last;
	}
    }
    return @dbservers;
}

sub usage( $ ) {
    my $ret = shift;
    print <<EOFusage;
Usage: $0 [-n|--noaction] [-v|--verbose]

Tells the AFS client which AFS servers in the local cell are closer
than the others, based on its own IPv4 address and hardcoded networks.
Unless "--noaction" is specified, this needs to be run as superuser.

EOFusage
exit($ret);
}
# main
GetOptions ("help" => \$opt_help,
	    "noaction" => \$opt_noaction,
	    "verbose" => \$opt_verbose) or usage(1);
if($opt_help) {
    usage(0);
}

if($> != 0 and not $opt_noaction) {
    print "#ERROR: needs to be run as root\n";
    exit(1);
}

my $cell = get_mycell();
openlog("cern_setserverprefs", "pid", LOG_LOCAL0);
my $own_ip = get_my_own_ip();
my $at_remote = ( ip_in_cidrs($own_ip, $remote_nets) );  # default:no.
if ($at_remote) {
   print "#INFO: will prefer AFS servers matching 'remote' networks\n" if $opt_verbose;
   syslog(LOG_INFO, "will prefer AFS servers matching 'close' networks");
} else {
   print "#INFO: will avoid AFS servers matching 'remote' networks\n" if $opt_verbose;
   syslog(LOG_INFO, "will avoid AFS servers matching 'remote' networks");
}

my $pref_fs_close=25000; # default is 30000 for same net, 40000 for other
my $pref_fs_remote=35000;
my @fsprefs;
for my $ip (get_fileservers($cell)) {
    if ((ip_in_cidrs($ip, $remote_nets)) == $at_remote) {
	# either both client+server at remote, or not => prefer
	push(@fsprefs,"$ip $pref_fs_close");
    } else {
	push(@fsprefs,"$ip $pref_fs_remote");
    }
}

my $pref_db_close=8000;  # default is 10000
my $pref_db_remote=12000;
my @dbprefs;
for my $ip (get_dbservers($cell)) {
    if ((ip_in_cidrs($ip, $remote_nets)) == $at_remote) {
	# either both client+server at remote, or not => prefer
	push(@dbprefs,"$ip $pref_db_close");
    } else {
	push(@dbprefs,"$ip $pref_db_remote");
    }
}

if($#fsprefs > 0) {
    my $cmd = "fs setserverprefs -servers ".join(" ", @fsprefs);
     if($opt_noaction) {
	print "# would run $cmd\n";
    } else {
	print "#INFO: running $cmd\n" if $opt_verbose;
	system($cmd);
    }
}
if($#dbprefs > 0) {
    my $cmd = "fs setserverprefs -vlservers ".join(" ", @dbprefs);
    if($opt_noaction) {
	print "# would run $cmd\n";
    } else {
	print "#INFO: running $cmd\n" if $opt_verbose;
	system($cmd);
    }
}
